#-------------------------------------------------------------------------------
# Name:        lac-sec
# Purpose:     evaluate the security of lac 
# Author:      Lu Xianhui
# Created:     18-09-2017
#-------------------------------------------------------------------------------
from math import sqrt, exp, log, pi, e
from decimal import *

# the root hermite factor delta of BKZ
def delta_bkz(b):
    return ((pi*b)**(1./b) * b / (2*pi*exp(1)))**(1./(2.*b-2.))

#the best plausible quantum cost of svp in dimension b
def svp_plausible(b):
    #return b*0.2057
    return b *log(sqrt(4./3))/log(2)

#the best known quantum cost of svp in dimension b
def svp_quantum(b):
    #return b*0.265
    return b *log(sqrt(13./9))/log(2)

#the classical cost of svp in dimension b
def svp_classical(b):
    #return b*0.292
    return b *log(sqrt(3./2))/log(2)

#the cost of sieve algorithm
def sieve(b):
    #return b*0.2075
    return b *log(sqrt(4./3))/log(2)


#the cost of primal attack
def cost_primal(n,q,s,cost_type):
    best_cost=10000
    best_b=0
    best_m=0
    loop=3*n+1
    for b in range(100,loop):
        if cost_type(b)>best_cost:
            break
        v_len=sqrt(b*s)
        delta=delta_bkz(b)
        for m in range(1,2*n):
            d=n+m+1
            gs_b_len=delta**(2.*b-d-1) * q**(1.*m /d)
            if v_len<gs_b_len:
                cost=cost_type(b)
                if cost<best_cost:
                    best_cost=cost
                    best_b=b
                    best_m=m

    return (best_m,best_b,int(best_cost))

#the cost of dual attack
def cost_dual(n,q,s,cost_type):
    loop=3*n
    best_cost=10000
    best_b=0
    best_m=0

    for b in range(100,loop):
        if cost_type(b)>best_cost:
            break
        delta=delta_bkz(b)
        for m in range(1,2*n):
            d=n+m
            l=delta**(d-1) * q**(1.*n/d)
            tau = l*sqrt(s)/q
            log2_eps = - 2 * pi * pi * tau**2 / log(2)
            log2_R = max( 0 , - 2 * log2_eps - sieve(b) )
            
            cost=cost_type(b)+log2_R
            if cost<best_cost:
                best_cost=cost
                best_b=b
                best_m=m

    return (best_m,best_b,int(best_cost))

#algebraic attack
def ag_cost(n,C):
    omiga=2
    beta=C-0.5-sqrt(C*(C-1))
    x=beta/(1+beta)
    H2=-x*log(x,2)-(1-x)*log(1-x,2)
    cost=omiga*n*(1+beta)*H2+log(n*n,2)
    return int(cost)
    #return int(0.344*n+log(n*n,2))

#test
def test_cost(n,q,s):

#root hermite factor
#    d=2*n+1.
#    sv=sqrt(2*n*s+1)
#    print "sv:",sv, "vol:",q**(n/d)
#    rf=sqrt(d/(2*pi*e))*(q**(n/d))/sv
#    print "Root Hermite Factor:", rf**(1./d)

#primal attack  cost
    print "primal attack:"
    cost_type=svp_classical
    (best_m,best_b,best_cost)=cost_primal(n,q,s,cost_type)
    print "classical: m=",best_m,"b=",best_b,"cost=",best_cost

    cost_type=svp_quantum
    (best_m,best_b,best_cost)=cost_primal(n,q,s,cost_type)
    print "quantum  : m=",best_m,"b=",best_b,"cost=",best_cost

    cost_type=svp_plausible
    (best_m,best_b,best_cost)=cost_primal(n,q,s,cost_type)
    print "plausilbe: m=",best_m,"b=",best_b,"cost=",best_cost
    
#dual attack cost
    print "dual attack:"
    cost_type=svp_classical
    (best_m,best_b,best_cost)=cost_dual(n,q,s,cost_type)
    print "classical: m=",best_m,"b=",best_b,"cost=",best_cost

    cost_type=svp_quantum
    (best_m,best_b,best_cost)=cost_dual(n,q,s,cost_type)
    print "quantum  : m=",best_m,"b=",best_b,"cost=",best_cost

    cost_type=svp_plausible
    (best_m,best_b,best_cost)=cost_dual(n,q,s,cost_type)
    print "plausilbe: m=",best_m,"b=",best_b,"cost=",best_cost

#algebraic attack
    print "algebraic attack cost:        ", ag_cost(n,6.6)


def main():
    
#set parameters
    q=251
   
    print "Concrete security of LAC128:"
    test_cost(512,q,1/2.)

    print "Concrete security of LAC192:"
    test_cost(1024,q,1/4.)

    print "Concrete security of LAC256:"
    test_cost(1024,q,1/2.)
    
    pass
if __name__ == '__main__':
    main()
